package main

import (
	"crypto/rand"
	"fmt"
	"os"
	"syscall"
	"testing"
)


// xpread is pread(2) but aborts on an error or when fewer bytes was read
func xpread(fd int, buf []byte, offset int64) {
	n, err := syscall.Pread(fd, buf, offset)
	if err != nil {
		panic(err)
	}
	if n != len(buf) {
		panic(fmt.Errorf("pread(%v) -> %v", len(buf), n))
	}
}

// xreadat is ReadAt but aborts on an error or when fewer bytes was read
func xreadat(f *os.File, buf []byte, offset int64) {
	n, err := f.ReadAt(buf, offset)
	if err != nil {
		panic(err)
	}
	if n != len(buf) {
		panic(fmt.Errorf("readat(%v) -> %v", len(buf), n))
	}
}

// setupXdat4K setups /dev/shm/x.dat with 4K content
func setupXdat4K(t interface {Fatal(...interface{})} /* = testing.TB */) *os.File {
	// setup 4K file on a tmpfs
	f, err := os.Create("/dev/shm/x.dat")
	if err != nil {
		t.Fatal(err)
	}

	buf4K := make([]byte, 4096)
	_, err = rand.Read(buf4K)
	if err != nil {
		// rand promises n = len(buf) <=> err == nil
		t.Fatal(err)
	}

	_, err = f.Write(buf4K)
	if err != nil {
		t.Fatal(err)
	}

	return f
}


// BenchmarkReadAt benchmarks os.File.ReadAt when reading 1 byte from a file on tmpfs
func BenchmarkReadAt1(b *testing.B) {
	f := setupXdat4K(b)
	buf1B := make([]byte, 1)

	// warm up
	xreadat(f, buf1B, 0)

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		xreadat(f, buf1B, 0)
	}
	b.StopTimer()

	err := f.Close()
	if err != nil {
		b.Fatal(f)
	}
}

// BenchmarkPread1 benchmarks pread syscall when reading 1 byte from a file on tmpfs
func BenchmarkPread1(b *testing.B) {
	f := setupXdat4K(b)
	fd := int(f.Fd())	// syscall wants fd as int, not uintptr
	buf1B := make([]byte, 1)

	// warm up
	xpread(fd, buf1B, 0)

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		xpread(fd, buf1B, 0)
	}
	b.StopTimer()

	err := f.Close()
	if err != nil {
		b.Fatal(f)
	}
}
